(* jpp.ml *)
(* Java Pretty Printer. Generates java code from a Java ast *)
(* @author: Lili Chen *)

open Jast

let default_keyword = "public static"
let default_implicit_param = "null"

let string_of_operator = function
	| Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
  | Equal -> "==" | Neq -> "!=" | Less -> "<" | Leq -> "<=" 
	| Greater -> ">" | Geq -> ">=" | Noop -> "" 
	| Not -> "!" | Inc -> "++" | Dec -> "--"
	
let rec string_of_expr = function
  | Literal(l) -> string_of_float l
	|	BooleanLiteral(s) -> s
  | Id(s) -> s
	| Unop(Not, e) -> "!" ^ "(" ^ string_of_expr e ^ ")"
	| Unop(op, e) -> "(" ^ string_of_expr e ^ ")" ^ string_of_operator(op)
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_operator(o) ^ " " ^ string_of_expr e2
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Noexpr -> "fix_me"
  | Call(f,f_Name,f_List) -> (if ((string_of_expr f) <> default_implicit_param) then ((string_of_expr f) ^ ".") else "" ) ^ 
			f_Name ^ "(" ^ String.concat "," (List.map string_of_expr f_List) ^ ")"
  | List(l) -> "[" ^ String.concat "," (List.map string_of_expr l) ^ "]"
  | StringLit(l) -> "\"" ^ l ^ "\""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(e) -> "return " ^ string_of_expr e ^ ";\n";
  | New(n,m,e_list) -> m ^ " = new " ^ n ^ "(" ^ String.concat "," (List.map string_of_expr e_list) ^ ");\n" 
	| While(e, s) -> "while (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
	| NewDecl(t,id,e_list) -> t ^ " " ^ id ^ " = new " ^ t ^ "(" ^ String.concat "," (List.map string_of_expr e_list) ^ ");\n"
	| Decl(t, id) -> t ^ " " ^ id ^ ";\n"
 (*  | If(e, s) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s *)
  | IfThenElse(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | Foreach(t, id, e, stmt) ->
      "for (" ^ t  ^ " " ^ id ^ " : " ^ string_of_expr e ^ ")\n" ^ string_of_stmt stmt ^ "\n"
    (* e.g. Foreach n : graph.getNode("a", "< 6", "visited") *)
	| Nostmt -> ""

let string_of_formal (t, id) = t ^ " " ^ id

let string_of_global vdecl = snd(vdecl) ^ " " ^ fst(vdecl) ^ ";"
 
let string_of_method_decl method_decl = 
	default_keyword ^ " " ^ method_decl.jrettype ^ " " ^ method_decl.jfname ^ "(" ^ 
	(String.concat "," (List.map string_of_formal method_decl.jformals))
	^ ")\n{\n\n" ^ 
	(String.concat "" (List.map string_of_stmt method_decl.jbody))
	^ "}\n"
      
(* class-name * globals-list * method-decls-list *) 
(* type java_class = string * string list * method_decl list *)
let string_of_jclass extra_string java_class =
  "\npublic class " ^ 
  (fun(x, y, z) -> x ^ "\n{\n" ^ String.concat "\n" (List.map string_of_global y) ^ "\n\n" ^ 
	String.concat "\n\n" (List.map string_of_method_decl z) ^ extra_string) java_class ^ "}\n\n"
  
(* e.g., import lib.*; *)	
let string_of_imports import_list = 
  String.concat ";\n" import_list ^ ";\n"
    
let string_of_program (import_list, java_class_list, extra_string) =
  "\n" ^ string_of_imports import_list ^ "\n" ^
  String.concat "\n\n" (List.map (string_of_jclass extra_string) java_class_list) ^ "\n\n\n"