(* translate.ml *)
(* Translates a GRAPL AST to a Java AST and performs semantic checking.*)
(* @authors Ryan Turner, Andres Uribe *)

type mode = Quiet | Verbose (* compiler directives *)

open Ast
open Jast

(* Function symbol table *)
module FuncMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

(* Exception types *)
exception ReturnException of string 
exception VariableNotFoundException of string 
exception FunctionNotFoundException of string 
exception BinopException of string
exception VisitUnvisitException of string
exception OtherException of string
exception AssignmentException of string
exception FunctionException of string
exception WhileExprException of string
exception ForeachNodeException of string
exception ForeachPredException of string
exception IfExprException of string
exception ReturnTypeException of string
exception MainNotFoundException of string
  
(* Default information for the backend java class that will run the GRAPL programs *)
(* let grapl_std_lib = GraplStdLib.string_of_library *)
let grapl_std_lib = "GraplStdLib"
let default_imports = ["lib.*"; grapl_std_lib] (* Java Backend classes *)
let default_lib = "library" (* Java Backend classes *)
let default_graph = "graph" (* Java Backend classes *)
let default_createGraph = "createEmptyGraph" (* Java Backend classes *)
let default_vars = [("GraplLib", default_lib); ("Graph", default_graph)] (* library must be created first *)
let default_keywords = "public static" 
let null_implicit_param = "null"
let main_formals = [("String[]", "args")]
let main_rettype = "void"
let default_prefix = "gpl_" (* prefix ensures that backend Java identifiers cannot conflict with the user's identifiers *) 
let java_globals_prefix = ""

(* Built-in functions are implemented in Java *)
let built_in_functions = [
	{ rettype = "number"; fname = "length"; formals = [("list", "l")]; locals = []; body = [] } ; 
	{ rettype = "node"; fname = "head"; formals = [("list", "l")]; locals = []; body = [] } ; 
	{ rettype = "list"; fname = "tail"; formals = [("list", "l")]; locals = []; body = [] } ; 
	{ rettype = "void"; fname = "unvisitAll"; formals = []; locals = []; body = [] } ;
	{ rettype = "boolean"; fname = "isVisited"; formals = [("node", "n")]; locals = []; body = [] } ;	 
	]

(* Library functions are written in GRAPL and defined in the GRAPL Standard Library. The user can extend *)
(* the standard library as needed and record the function definitions here. *)
let library_functions = [
	{ rettype = "number"; fname = "numChildren"; formals = [("node", "n")]; locals = []; body = [] } ; 
	{ rettype = "number"; fname = "numAncestors"; formals = [("node", "n")]; locals = []; body = [] } ; 
	{ rettype = "void"; fname = "unvisitAllFrom"; formals = [("node", "n")]; locals = []; body = [] } ; 
	{ rettype = "void"; fname = "visitAllFrom"; formals = [("node", "n")]; locals = []; body = [] } ; 
	{ rettype = "list"; fname = "dfs"; formals = [("node", "start"); ("node", "finish")]; locals = []; body = [] } ; 
	{ rettype = "list"; fname = "reverse"; formals = [("list", "old_lidt")]; locals = []; body = [] } ; 

	]

let libTable_data = (* Put standard library function declarations in a symbol table *)
			List.fold_left (fun funcs fdecl -> FuncMap.add fdecl.fname fdecl funcs)
      FuncMap.empty (library_functions)
			
let initialize_standard_library = Jast.Expr(Jast.Call(Jast.Id(grapl_std_lib), "initialize", [Jast.Id(default_graph); Jast.Id(default_lib)]))

(* Functions to extract the correct data structure from env tuple *)
(* env = (var_list, sp, funcTable, libTable, mode) *)
let var_list (v, s, f, l, m) = v
let sp (v, s, f, l, m) = s
let funcTable (v, s, f, l, m) = f
let libTable (v, s, f, l, m) = l
let mode (v, s, f, l, m) = m

(* Utility functions for pushing and popping local vars into symbol table var_list *)
(* var_list is the symbol table; sp keeps track of how many locals are currently in scope *)
let push_symbols env vars = 
	let var_list = ((List.append vars (var_list env))) 
	and sp = ((List.length vars) :: (sp env))
	in (var_list, sp, funcTable env, libTable env, mode env)

let pop sp = ((List.hd sp)-1)::(List.tl sp)
let rec pop_sym var_list sp = match var_list with
  | []-> [] (* failwith "empty" *)
  | hd::tl-> if(List.hd sp)>0 then pop_sym tl (pop sp) else hd::tl
let pop_symbols env = ((pop_sym (var_list env) (sp env)), (List.tl (sp env)), funcTable env, libTable env, mode env)

(* Other utility functions *)
let flip_append_list string_tuple_list = List.map (fun (t, id) -> (default_prefix ^ id, t)) string_tuple_list
let flip (x, y) = (y, x)
let flip_list string_tuple_list = List.map flip string_tuple_list
let rec string_of_var_list var_list = 
	let string_of_tuple (t, id) = "(" ^ t ^ "," ^ id ^ ")" in
	string_of_tuple (List.hd var_list) ^ (if (List.length var_list = 1) then "" else string_of_var_list (List.tl var_list))
let append_keywords_list string_tuple_list = List.map (fun (id, t) -> (id, default_keywords ^ " " ^ t)) string_tuple_list

let java_type_for_type t = match t with
	| "node" -> "Node"
	| "graph" -> "Graph"
	| "number" -> "double"
	| "list" -> "List"
	| "void" -> "void"
	| "boolean" -> "boolean"
	| _  -> ""

let java_operator_for_operator o = match o with
	| Ast.Add -> Jast.Add
	| Ast.Sub -> Jast.Sub
	| Ast.Mult -> Jast.Mult
	| Ast.Div -> Jast.Div
	| Ast.Equal -> Jast.Equal
	| Ast.Neq -> Jast.Neq
	| Ast.Greater -> Jast.Greater
	| Ast.Geq -> Jast.Geq
	| Ast.Less -> Jast.Less
	| Ast.Leq -> Jast.Less
	| Ast.Noop -> Jast.Noop
	| Ast.Cons -> Jast.Noop
	| Ast.Not -> Jast.Not
	| Ast.Inc -> Jast.Inc
	| Ast.Dec -> Jast.Dec 

let node_name_for_string s = default_prefix ^ s

(***********************************************************************************************************************)
(* SEMANTIC CHECKING *)

let print_warning mode w = match mode with 
	| Verbose -> print_string("/* WARNING: " ^ w ^ "*/\n")
	| Quiet -> print_string ""
		
(* Translating functions *)

(* Returns the type for a grapl expression, or raises exception for an invalid type. *)
let rec get_type_for env e = match e with
	| Ast.Literal(l) -> "number"
	| Ast.BooleanLiteral(s) -> "boolean"
	| Ast.Call(f, e_list) -> 
   	 (let fdecl = try FuncMap.find f (funcTable env) 
                 with Not_found -> raise(FunctionNotFoundException(f ^" has not been declared"))
      in fdecl.rettype)
	| Ast.Noexpr -> "noexpr"
	| Ast.Unvisit(e) -> "node"
	| Ast.Visit(e) -> "node"
	| Ast.List(l) -> "list"	
	| Ast.Unop(op, e) -> get_type_for env e
	| Ast.Binop(e1, op, e2) -> (match op with 
   		| Ast.Equal | Ast.Neq | Ast.Less | Ast.Leq | Ast.Greater | Ast.Geq -> "boolean"
                                                              | _ -> get_type_for env e2)
	| Ast.Assign(v, e) -> get_type_for env e

	(* Implicit global declaration of nodes inside the "graph" statement means that nodes are created dynamically. The compiler *)
	(* cannot guarantee in all cases that a reference to an implicitly declared node will be valid; thus a warning is generated. *)
	| Ast.Id(s) ->   
	   (try (List.assoc s (var_list env)) 
			with Not_found -> print_warning (mode env) ("reference to implicitly declared node \'" ^ s ^ "\' may not be valid at run-time."); "node" )

(* returns boolean after checking *)
let rec check_expr env e = 
	match e with
    | Ast.Id(s) -> ((List.mem_assoc s (var_list env)) || ((get_type_for env (Ast.Id(s))) = "node"))
		| Ast.Unop(Ast.Not, e) -> get_type_for env e = "boolean"
		| Ast.Unop(op, e) -> get_type_for env e = "number" 
		| Ast.Binop(e1,o,e2) -> 
      if (check_expr env e1 && check_expr env e2) 
			then
	  		(match o with 
	  			| Ast.Cons -> get_type_for env e1 = "node" && get_type_for env e2 = "list"													
	  			| _  -> ((get_type_for env e1) = (get_type_for env e2)) 
				) 
			else 
				false
    | Ast.Assign(v,e) -> if (check_expr env e) 
				then 
					(get_type_for env (Ast.Id(v))) = (get_type_for env e)
				else
					false
    | Ast.Call("print", e_list) -> true
      
    | Ast.Call(f, e_list) -> (FuncMap.mem f (funcTable env)) || (FuncMap.mem f (libTable env)) (* TODO: check argument list *) 
    | Ast.List(e_list) -> List.fold_left (fun a e -> a && (check_expr env e)) true e_list
    | Ast.Literal(f) -> true
		| Ast.BooleanLiteral(s) -> true
    | Ast.Unvisit(n) -> true
    | Ast.Visit(n) -> true
    | Ast.Noexpr  -> true

(* Raises an exception for an invalid expression *)			      
let consistent_expr env e =
  let inner = check_expr env e
  in
  if (inner) then true
  else match e with
    | Ast.Id(s) -> raise(VariableNotFoundException("The variable \'" ^ s ^ "\' was not declared or is out of scope."))
    | Ast.Binop(e1,op,e2) -> raise(BinopException("Problem with the binary expression: \'" ^ Gpp.string_of_expr(Ast.Binop(e1,op,e2)) ^ "\'."))
    | Ast.Unop(op,e2) -> raise(BinopException("Problem with the unary expression: \'" ^ Gpp.string_of_expr(Ast.Unop(op,e)) ^ "\'."))
		| Ast.Assign(v,e1) -> raise(AssignmentException("Type error: \'" ^ v ^ "\' cannot be assigned to " ^ Gpp.string_of_expr(e1) ^ "\'."))
    | Ast.Call(f, e_list) -> raise(FunctionNotFoundException("The function \'"^ f ^"\' has not been declared."))
    | Ast.List(e_list) -> raise(OtherException("Undefined exception in list."))
    | Ast.Literal(f) -> raise(OtherException("Undefined exception in literal \'" ^ string_of_float f ^ "\'."))
    | Ast.BooleanLiteral(s) -> raise(OtherException("Undefined exception in boolean \'" ^ s ^ "\'."))
		| Ast.Unvisit(n) -> raise(OtherException("Cannot call function 'unvisit' on non-node \'" ^ n ^ "\'."))
    | Ast.Visit(n) -> raise(OtherException("Cannot call function 'visit' on non-node \'" ^ n ^ "\'."))
    | Ast.Noexpr  -> raise(OtherException("Undefined exception in Noexpr."))
 

(***********************************************************************************************************************)
(* CONVERTING AST --> JAST (Java-Ast) *)

(* Returns a Jast.expr for a given Ast.expr *)										
let rec jast_of_expr env e =
	let jast_e expr = jast_of_expr env expr in
	if not(consistent_expr env e) then
		Jast.Noexpr (* invalid expressions should throw exceptions before reaching this point *)
	else		 	
		match e with
		| Ast.Literal(l) -> Jast.Literal(l)
		| Ast.BooleanLiteral(s) -> Jast.BooleanLiteral(s)
	  | Ast.Id(s) -> 
				if (not(List.mem_assoc s (var_list env)) && (get_type_for env (Ast.Id(s))) = "node") 
				then Jast.Call(Jast.Id(default_graph), "getNode", [Jast.StringLit(s)]) 
				else Jast.Id(default_prefix ^ s) 
		| Ast.Unop(op, e) -> Jast.Unop(java_operator_for_operator op, jast_e e)		
		| Ast.Binop(e1, Ast.Cons, e2) -> Jast.Call(jast_e e2, "addNew", [jast_e e1])
	  | Ast.Binop(e1, op, e2) -> Jast.Binop(jast_e e1, (java_operator_for_operator op), jast_e e2)
	  | Ast.Assign(v, e) -> Jast.Assign(default_prefix ^ v, jast_e e)
	  | Ast.Noexpr ->  Jast.Noexpr
	  | Ast.Unvisit(e) -> Jast.Call(jast_e (Ast.Id(e)), "unvisit", []) 
	  | Ast.Visit(e) ->  Jast.Call(jast_e (Ast.Id(e)), "visit", []) 
	  | Ast.List(l) -> Jast.Call(Jast.Id(default_lib), "buildList", List.map jast_e l)
	
		(* Built-in functions *)
	  | Ast.Call("print", e_list) -> Jast.Call(Jast.Id(default_lib), "print", List.map jast_e e_list)
	 	| Ast.Call("unvisitAll", e_list) -> Jast.Call(Jast.Id(default_graph), "unvisitAll", [])
		| Ast.Call("length", e_list) -> let e = List.hd e_list in let x = get_type_for env e in
				if (x = "list") 
				then Jast.Call(jast_e e, "size", []) 
				else raise (FunctionException("Error: Cannot call built-in 'length' on non-list \'" ^ Gpp.string_of_expr(e) ^ "\'."))  
		| Ast.Call("head", e_list) -> let e = List.hd e_list in let x = get_type_for env e in
				if (x = "list") 
				then Jast.Call(Jast.Id(default_lib), "head", [jast_e e]) 
				else raise (FunctionException("Error: Cannot call built-in 'head' on non-list \'" ^ Gpp.string_of_expr(e) ^ "\'."))
		| Ast.Call("tail", e_list) -> let e = List.hd e_list in let x = get_type_for env e in
				if (x = "list") 
				then Jast.Call(Jast.Id(default_lib), "tail", [jast_e e]) 
				else raise (FunctionException("Error: Cannot call built-in 'tail' on non-list \'" ^ Gpp.string_of_expr(e) ^ "\'."))  			
		| Ast.Call("isVisited", e_list) -> let e = List.hd e_list in let x = get_type_for env e in
				if (x = "node") 
				then Jast.Call(Jast.Id(default_lib), "isVisited", [jast_e e]) 
				else raise (FunctionException("Error: Cannot call built-in 'isVisited' on non-node \'" ^ Gpp.string_of_expr(e) ^ "\'."))  			

		(* Other functions. Check the locally-declared functions first; this way, the user can override library functions if desired.*)
		(* If the function is not found in the local function table, it must be a library function (since it is a valid expression). *)
		| Ast.Call(f, e_list) -> 
				if (FuncMap.mem f (funcTable env)) 
				then Jast.Call(Jast.Id(null_implicit_param), f, List.map jast_e e_list) (* user-defined functions *)
				else Jast.Call(Jast.Id(grapl_std_lib), f, List.map jast_e e_list) (* standard library functions *)

(* Returns a function call to create a new edge in the graph *)
let jast_of_edge_def edge_def = match edge_def with
	| (s1, Ast.Redge(w), s2) ->  
			Jast.Expr(Jast.Call(Jast.Id("graph"), "addEdge", [Jast.StringLit(s1); Jast.StringLit(s2); Jast.Literal(w)]))
	| (s1, Ast.Ledge(w), s2) -> 
			Jast.Expr(Jast.Call(Jast.Id("graph"), "addEdge", [Jast.StringLit(s2); Jast.StringLit(s1); Jast.Literal(w)]))
	| (s1, Ast.Bedge(w), s2) -> 
			Jast.Expr(Jast.Call(Jast.Id("graph"),"addBEdge", [Jast.StringLit(s1); Jast.StringLit(s2); Jast.Literal(w)])) 
	| (s1, Ast.NoEdge(w), s2) -> Jast.Nostmt

(* Returns a Jast.stmt list ** NOT a Jast.stmt ** for each Ast.stmt *)
let rec jast_of_stmt rettype env stmt = 
	let jast_s s = jast_of_stmt rettype env s in
	(*let jfold s_list = List.fold_left (jast_of_stmt env) s_list in*)
	let rec jast_of_foreach env (q, n1, ft, n2, (op,e), s) = 
		let env = push_symbols env [(n1, "node")] in
			let jast_e e = jast_of_expr env e in 
			match ft with
				| "from" -> 
						[Jast.Foreach("Node",
													default_prefix ^ n1,
													Jast.Call(Jast.Id("graph"),
							 											"getNodesFromWith",
							 											[ (jast_e (Ast.Id(n2)));
																		  Jast.Id("null");
																			Jast.StringLit(q);
																			Jast.StringLit(Gpp.string_of_operator(op));
																			jast_of_expr env e
																		]),
													Jast.Block(jast_of_stmt rettype env s)
													)
						]
				| "to" -> 
						[Jast.Foreach("Node",
													default_prefix ^ n1,
													Jast.Call(Jast.Id("graph"),
							 											"getNodesToWith",
							 											[ (jast_e (Ast.Id(n2)));
																		  Jast.Id("null");
																			Jast.StringLit(q);
																			Jast.StringLit(Gpp.string_of_operator(op));
																			jast_of_expr env e
																		]),
													Jast.Block(jast_of_stmt rettype env s)
													)
						]
						
				| _ -> [Jast.Nostmt]
	in	
		let jast_e e = jast_of_expr env e in 
			match stmt with 
		| Ast.Expr(e) -> [Jast.Expr(jast_e e)]			
		| Ast.Graph(edge_defs) -> List.map jast_of_edge_def edge_defs 
		| Ast.Block(stmts) -> [Jast.Block(List.fold_left (fun a b -> a @ b) [] (List.map (jast_of_stmt rettype env) stmts ))]
		| Ast.Return(e) -> if (not(rettype = get_type_for env e))
  			then raise(ReturnTypeException("Error: Return type of the function doesn't match with the return type in \'"^Gpp.string_of_expr e^"\'."))
            else [Jast.Return(jast_e e)]
	  | Ast.IfThenElse(e, s1, s2) ->  
	    if (not(  get_type_for env e = "boolean") )
	    then raise(IfExprException("Error: Expresion inside the if clause must evaluate to a boolean in \'"^ Gpp.string_of_expr e^"\'."))
	    else [Jast.IfThenElse(jast_e e, Jast.Block(jast_s s1), Jast.Block(jast_s s2))]
	  | Ast.Foreach(q,n1,ft,n2,(o,e),s) -> 
      if(not( get_type_for env (Ast.Id(n2)) = "node"))
      then raise(ForeachNodeException("Error: Identifier after the from/to has to resolve to a node type." ))
      else if (not( get_type_for env e = "number" ))
     	then raise(ForeachPredException ("Error: Predicate expression: \'"^Gpp.string_of_expr e^"\'. doesn't evaluate to a number inside foreach loop"))
      else jast_of_foreach env (q, n1, ft, n2, (o,e), s)
   	| Ast.While(e,s) -> 
     (* type checking *)
     if (get_type_for env e = "boolean") then
       [Jast.While(jast_e e, Jast.Block(jast_s s))]
     else 
       raise(WhileExprException("Error: Expresion inside the while loop has to evaluate to a boolean in \'"^Gpp.string_of_expr e^"\'."))

(* Works on Java types. Java primitives do not need to be instantiated; objects do. *)
let instantiate prefix (id, t) = match t with
	| "Node" -> Jast.New(t, prefix ^ id, [Jast.StringLit(id)])
	| "List" -> Jast.New(t, prefix ^ id, [])
	| "GraplLib" -> Jast.New(t, prefix ^ id, [])
	| "Graph" -> Jast.Expr(Jast.Assign(default_graph, Jast.Call(Jast.Id(default_lib), default_createGraph, [])))
	| _  -> Jast.Decl(java_type_for_type t, prefix ^ id)

(* Works on grapl types only *)
let declare_and_instantiate prefix (t, id) = match t with
	| "node" -> Jast.NewDecl(java_type_for_type t, prefix ^ id, [Jast.StringLit(id)])
	| "list" -> Jast.NewDecl(java_type_for_type t, prefix ^ id, [])
	| _  -> Jast.Decl(java_type_for_type t, prefix ^ id)

(* Returns a Jast.method_decl for each function in the grapl file *)		
let jast_of_function globals_list env func =
		{   
			jrettype = if func.fname = "main" then main_rettype else java_type_for_type func.rettype;
			jfname = func.fname;
			jformals = 
				if func.fname = "main" 
				then main_formals
   			else List.map (fun (t, id) -> (java_type_for_type t, default_prefix ^ id) ) func.formals;
			jlocals = List.map (fun (t, id) -> (java_type_for_type t, default_prefix ^ id) ) func.locals;
    		(* push extra initializations when doing the main *)
			jbody = 
				let env = push_symbols env ((List.map flip func.locals) @ (List.map flip func.formals))
				in 
    		if (func.fname = "main") 
				then  
					(List.map (instantiate java_globals_prefix) globals_list) @ 
					(List.map (declare_and_instantiate default_prefix) func.locals) @
					([initialize_standard_library]) @
					(List.fold_left (fun a b -> a @ b) [] (List.map (jast_of_stmt func.rettype env) func.body))
				else
					(List.map (declare_and_instantiate default_prefix) func.locals) @
					(List.fold_left (fun a b -> a @ b) [] (List.map (jast_of_stmt func.rettype env) func.body))   
			} 

(* Where the magic happens *)
let translate vars funcs program_name mode_data =
	let ft = (* Put function declarations in a symbol table *)
			(List.fold_left (fun funcs fdecl -> FuncMap.add fdecl.fname fdecl funcs)
      FuncMap.empty (funcs @ built_in_functions) )
	in let env = ([], [], ft, libTable_data, mode_data)
	in let env = push_symbols env (List.map flip vars) (* returns updated variable symbol table & stack-pointer list *)
	in let jast_of_program env globals_list func_list = 
		
			(* Java program is a tuple (imports-list, class-list, default_lib) *)
			let globals = (flip_append_list (List.map (fun (t,id) -> (java_type_for_type t, id)) globals_list)) @ (flip_list default_vars)
			in
   			if not(FuncMap.mem "main" (funcTable env)) 
      	then raise(MainNotFoundException("Error: main function was not found in the input file"))
   			else [(program_name, append_keywords_list globals, List.map (jast_of_function globals env) func_list)]
			
 in jast_of_program env vars funcs
